
import logging

from colibris import settings
from colibris import utils
from colibris.cache.base import ABSENT


DEFAULT_LIFETIME = 300  # seconds

logger = logging.getLogger(__name__)

_backend = None
_default_lifetime = None


def get_backend():
    global _backend, _default_lifetime

    if _backend is None:
        backend_settings = dict(settings.CACHE)
        backend_path = backend_settings.pop('backend', 'colibris.cache.locmem.LocMemBackend')
        backend_class = utils.import_member(backend_path)
        _default_lifetime = backend_settings.pop('default_lifetime', DEFAULT_LIFETIME)
        _backend = backend_class(**backend_settings)

        logger.debug('backend of class %s created', backend_path)

    return _backend


def get(key, default=None):
    value = get_backend().get(key)
    if value is ABSENT:
        return default

    return value


def set(key, value, lifetime=None):
    lifetime = lifetime or _default_lifetime

    get_backend().set(key, value, lifetime)


def delete(key):
    get_backend().delete(key)




import logging
import os

from colibris import settings
from colibris import utils

from .exceptions import *


PACKAGE_TEMPLATE_PATH = 'templates'

logger = logging.getLogger(__name__)

_backend = None


def get_backend():
    global _backend

    if _backend is None:
        backend_settings = dict(settings.TEMPLATE)
        backend_path = backend_settings.pop('backend', None)
        if backend_path is None:
            raise TemplateNotConfigured()

        # inject package dir into path array
        paths = backend_settings.get('paths', [])
        paths = [os.path.join(settings.PROJECT_PACKAGE_DIR, PACKAGE_TEMPLATE_PATH)] + list(paths)
        backend_settings['paths'] = paths

        backend_class = utils.import_member(backend_path)
        _backend = backend_class(**backend_settings)

        logger.debug('backend of class %s created', backend_path)

    return _backend


def render(template_name, **context):
    return get_backend().render(template_name, context)


def render_string(template_str, **context):
    return get_backend().render_string(template_str, context)




import logging

from colibris import settings
from colibris import utils
from colibris.email.message import EmailMessage


logger = logging.getLogger(__name__)

_backend = None


def get_backend():
    global _backend

    if _backend is None:
        backend_settings = dict(settings.EMAIL)
        backend_path = backend_settings.pop('backend', 'colibris.email.base.NullBackend')
        backend_class = utils.import_member(backend_path)

        logger.debug('creating backend of class %s', backend_path)

        _backend = backend_class(**backend_settings)

    return _backend


def send(email_message):
    logger.debug('sending %s', email_message)
    return get_backend().send_messages([email_message])


def send_many(email_messages):
    logger.debug('sending %d messages', len(email_messages))
    return get_backend().send_messages(email_messages)





class TaskQueueException(Exception):
    pass


class TaskQueueNotConfigured(TaskQueueException):
    def __init__(self):
        super().__init__('task queue mechanism has not been configured')


class TimeoutException(TaskQueueException):
    def __init__(self, timeout):
        super().__init__('task did not complete within {} seconds'.format(timeout))


class UnpicklableException(TaskQueueException):
    def __init__(self, exc_str):
        self.exc_str = exc_str

    def __str__(self):
        return self.exc_str




import logging
import sys

from colibris.email.base import EmailBackend


logger = logging.getLogger(__name__)


class ConsoleBackend(EmailBackend):
    def __init__(self, **kwargs):
        self.stream = sys.stdout

        super().__init__(**kwargs)

    def send_messages(self, email_messages):
        for message in email_messages:
            prepared = message.prepare()
            data = prepared.as_bytes()
            text = data.decode()

            self.stream.write('-' * 10 + '\n')
            self.stream.write(text)

        self.stream.write('-' * 10 + '\n')




from aiohttp import web

from colibris import settings
from colibris import app

from . import BaseCommand


class RunServerCommand(BaseCommand):
    def execute(self, options):
        web.run_app(app.webapp, host=settings.LISTEN, port=settings.PORT)




import marshmallow_peewee.schema

from marshmallow import Schema, ValidationError
from marshmallow import pre_dump, post_dump, pre_load, post_load, validates_schema
from marshmallow import fields, validate

from colibris.api import envelope
from colibris.utils import camelcase_to_underscore


DEFAULT_DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S'


class ModelSchemaOpts(marshmallow_peewee.schema.SchemaOpts):
    def __init__(self, meta, **kwargs):
        super().__init__(meta, **kwargs)

        if self.model:
            self.name = getattr(meta, 'name', camelcase_to_underscore(self.model.__name__))
            self.name_plural = getattr(meta, 'name_plural', self.name + 's')

        self.load_instance = getattr(meta, 'load_instance', False)

        self.datetimeformat = DEFAULT_DATETIME_FORMAT


class ModelSchema(marshmallow_peewee.ModelSchema):
    OPTIONS_CLASS = ModelSchemaOpts

    @post_dump(pass_many=True)
    def wrap(self, data, many):
        if many:
            return envelope.wrap_many(data)

        else:
            return envelope.wrap_one(data)

    @post_load
    def make_instance(self, data):
        if not self.opts.load_instance:
            return data

        return super().make_instance(data)


def many_envelope(schema_class):
    class SchemaWrapper(Schema):
        results = fields.Nested(schema_class, many=True)
        count = fields.Integer()
        pages = fields.Integer()
        page = fields.Integer()
        page_size = fields.Integer()

    return SchemaWrapper()




import logging

from colibris import settings
from colibris import utils

from .exceptions import *


_REQUEST_ACCOUNT_ITEM_NAME = 'account'


logger = logging.getLogger(__name__)

_backend = None


def get_backend():
    global _backend

    if _backend is None:
        backend_settings = dict(settings.AUTHENTICATION)
        backend_path = backend_settings.pop('backend', 'colibris.authentication.base.NullBackend')
        backend_class = utils.import_member(backend_path)

        logger.debug('creating backend of class %s', backend_path)

        _backend = backend_class(**backend_settings)

    return _backend


def authenticate(request):
    account = get_backend().authenticate(request)
    request[_REQUEST_ACCOUNT_ITEM_NAME] = account

    return account


def get_account(request):
    return request.get(_REQUEST_ACCOUNT_ITEM_NAME)


def login(request, account, persistent):
    logger.debug('logging in account "%s" (persistent=%s)', account, persistent)

    request[_REQUEST_ACCOUNT_ITEM_NAME] = account
    get_backend().login(request, persistent)


def logout(request):
    account = request.pop(_REQUEST_ACCOUNT_ITEM_NAME, None)
    if account:
        logger.debug('logging out account "%s"', account)
        get_backend().logout(request)


def process_request(request):
    return get_backend().process_request(request)


def process_response(request, response):
    return get_backend().process_response(request, response)




from colibris import utils
from colibris.authentication.base import AuthenticationBackend


class ModelBackend(AuthenticationBackend):
    def __init__(self, model, identity_field, secret_field, active_field=None, inactive_field=None, **kwargs):
        self.model = utils.import_member(model)
        self.identity_field = identity_field
        self.secret_field = secret_field
        self.active_field = active_field
        self.inactive_field = inactive_field

        super().__init__(**kwargs)

    def lookup_account(self, identity):
        query = self.get_identity_field() == identity
        if self.active_field:
            query = query & (self.get_active_field() == True)

        if self.inactive_field:
            query = query & (self.get_inactive_field() == False)

        try:
            return self.model.select().where(query).get()

        except self.model.DoesNotExist:
            return None

    def get_identity_field(self):
        return getattr(self.model, self.identity_field)

    def get_active_field(self):
        return getattr(self.model, self.active_field)

    def get_inactive_field(self):
        return getattr(self.model, self.inactive_field)

    def get_secret(self, account):
        return getattr(account, self.secret_field)




import pickle
import redis

from colibris.cache.base import CacheBackend, ABSENT


class RedisBackend(CacheBackend):
    def __init__(self, host='localhost', port=6379, db=0, password=None):
        self._host = host
        self._port = port
        self._db = db

        self._client = redis.Redis(host=host, port=port, db=db, password=password)

    def get(self, key):
        svalue = self._client.get(key)
        if svalue is None:
            return ABSENT

        return self._load_value(svalue)

    def set(self, key, value, lifetime):
        self._client.setex(key, lifetime, self._dump_value(value))

    def delete(self, key):
        self._client.delete(key)

    def _dump_value(self, value):
        return pickle.dumps(value)

    def _load_value(self, svalue):
        return pickle.loads(svalue)




import aiosmtplib
import asyncio
import logging

from colibris.email.base import EmailBackend


DEFAULT_TIMEOUT = 60

logger = logging.getLogger(__name__)


class SMTPBackend(EmailBackend):
    def __init__(self, host, port, username=None, password=None, use_tls=False, timeout=DEFAULT_TIMEOUT, **kwargs):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.use_tls = use_tls
        self.timeout = timeout

        super().__init__(**kwargs)

    def send_messages(self, email_messages):
        asyncio.ensure_future(self.send_messages_async(email_messages))

    async def send_messages_async(self, email_messages):
        logger.debug('connecting to %s:%d', self.host, self.port)
        try:
            client = await self.connect()

        except Exception as e:
            logger.error('failed to connect: %s', e, exc_info=True)
            return

        for i, message in enumerate(email_messages):
            logger.debug('sending message %d/%d', i + 1, len(email_messages))
            try:
                await client.send_message(message.prepare())

            except Exception as e:
                logger.error('failed to send message (%s): %s', message, e, exc_info=True)

        logger.debug('closing connection')
        client.close()

    async def connect(self):
        client = aiosmtplib.SMTP(timeout=self.timeout)
        await client.connect(hostname=self.host, port=self.port)
        if self.use_tls:
            await client.starttls()

        if self.username:
            await client.login(self.username, self.password)

        return client




import logging


logger = logging.getLogger(__name__)


class EmailBackend:
    def __init__(self, **kwargs):
        pass

    def send_messages(self, email_messages):
        raise NotImplementedError


class NullBackend(EmailBackend):
    def send_messages(self, email_messages):
        logger.debug('null backend: send_messages() called')




import mimetypes
import socket

from email import encoders
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formatdate, make_msgid

from colibris import settings


_cached_domain = None


class EmailMessage:
    def __init__(self, subject, body, to, cc=None, bcc=None,
                 from_=None, reply_to=None, headers=None, html=None):

        self.subject = subject
        self.body = body

        self.from_ = from_ or settings.EMAIL['default_from']
        self.to = to
        self.cc = cc
        self.bcc = bcc
        self.reply_to = reply_to

        self.headers = headers or {}  # TODO headers should be case insensitive
        self.attachments = []
        self.html = html

    def __str__(self):
        to = ', '.join(('<{}>'.format(t) for t in self.to))
        return 'email to {to}: {subject}'.format(to=to, subject=self.subject)

    def prepare(self):
        msg = MIMEText(self.body, _charset='utf-8')
        msg = self._prepare_attachments(msg)
        msg = self._prepare_headers(msg)

        return msg

    def attach(self, filename, content, mimetype=None):
        mimetype = mimetype or mimetypes.guess_type(filename)[0]
        basetype, subtype = mimetype.split('/', 1)

        if basetype == 'text' and isinstance(content, bytes):
            content = content.decode()

        self.attachments.append((content, mimetype, filename))

    def _prepare_headers(self, msg):
        global _cached_domain

        msg['Subject'] = self.subject
        msg['From'] = self.headers.get('From', self.from_)
        msg['To'] = ', '.join(self.headers.get('To', self.to))

        cc = self.headers.get('Cc')
        if cc is None and self.cc:
            cc = ', '.join(self.cc)
        if cc is not None:
            msg['Cc'] = cc

        bcc = self.headers.get('Bcc')
        if bcc is None and self.bcc:
            bcc = ', '.join(self.bcc)
        if bcc is not None:
            msg['Bcc'] = bcc

        reply_to = self.headers.get('Reply-To')
        if reply_to is None and self.reply_to:
            reply_to = ', '.join(self.reply_to)
        if reply_to is not None:
            msg['Reply-To'] = reply_to

        date = self.headers.get('Date')
        if date is None:
            date = formatdate()
        msg['Date'] = date

        msg_id = self.headers.get('Message-ID')
        if msg_id is None:
            if _cached_domain is None:
                _cached_domain = socket.getfqdn()

            msg_id = make_msgid(domain=_cached_domain)
        msg['Message-ID'] = msg_id

        for name, value in self.headers.items():
            msg[name] = value

        return msg

    def _prepare_attachments(self, msg):
        if self.html:
            self.attachments.append((self.html, 'text/html', None))

        if self.attachments:
            subtype = 'mixed'
            if self.html:
                # If html is supplied, use it as an alternative to textual body
                subtype = 'alternative'

            multipart_msg = MIMEMultipart(_subtype=subtype)
            multipart_msg.attach(msg)

            for attachment in self.attachments:
                multipart_msg.attach(self._prepare_attachment(*attachment))

            return multipart_msg

        return msg

    def _prepare_attachment(self, content, mimetype, filename=None):
        basetype, subtype = mimetype.split('/', 1)
        if basetype == 'text':
            attachment = MIMEText(content, subtype, _charset='utf-8')

        else:
            attachment = MIMEBase(basetype, subtype)
            attachment.set_payload(content)
            encoders.encode_base64(attachment)

        if filename:
            try:
                filename.encode('ascii')

            except UnicodeEncodeError:
                filename = ('utf-8', '', filename)

            attachment.add_header('Content-Disposition', 'attachment', filename=filename)

        return attachment




import json
import logging
import re

from aiohttp import web
from aiohttp_apispec import validation_middleware
from webargs import aiohttpparser

from colibris import api
from colibris import app
from colibris import authentication
from colibris import authorization
from colibris import settings
from colibris import utils
from colibris.api import BaseJSONException, envelope


logger = logging.getLogger(__name__)


class HTTPSchemaValidationError(web.HTTPUnprocessableEntity):
    def __init__(self, schema_error, **kwargs):
        super().__init__(**kwargs)

        self.details = schema_error.messages


def _extract_request_logging_info(request):
    info = '{method} {path}'.format(method=request.method, path=request.path)
    account = authentication.get_account(request)
    if account:
        info += ' (account={})'.format(account)

    info += ': '

    return info


@web.middleware
async def handle_errors_json(request, handler):
    try:
        return await handler(request)

    except BaseJSONException as e:
        logger.warning('%s%s', _extract_request_logging_info(request), e)

        return web.json_response(envelope.wrap_error(e.code, e.message, e.details), status=e.status)

    except HTTPSchemaValidationError as e:
        code = 'invalid_fields'
        message = 'Some of the supplied fields are invalid.'
        logger.warning('%sschema validation error: %s', _extract_request_logging_info(request), e.details)

        return web.json_response(envelope.wrap_error(code, message, e.details), status=e.status)

    except web.HTTPRedirection as e:
        return e

    except (web.HTTPClientError, web.HTTPServerError) as e:
        code = utils.camelcase_to_underscore(re.sub('[^a-zA-Z0-9_]', '', e.reason))
        message = e.reason
        logger.warning('%s%s', _extract_request_logging_info(request), e)

        return web.json_response(envelope.wrap_error(code, message), status=e.status)

    except json.JSONDecodeError as e:
        code = 'invalid_json'
        message = 'Invalid JSON.'
        details = str(e)
        logger.warning('%sinvalid JSON: %s', _extract_request_logging_info(request), e)

        return web.json_response(envelope.wrap_error(code, message, details), status=400)

    except app.HealthException as e:
        code = 'unhealthy'
        message = 'Service is not healthy.'
        details = str(e)
        logger.critical('%sservice is not healthy: %s', _extract_request_logging_info(request), e)

        return web.json_response(envelope.wrap_error(code, message, details), status=500)

    except Exception as e:
        code = 'server_error'
        message = 'Internal Server Error'
        details = None
        logger.error('%sinternal server error: %s', _extract_request_logging_info(request), e, exc_info=True)

        if settings.DEBUG:
            import traceback

            details = traceback.format_exc()

        return web.json_response(envelope.wrap_error(code, message, details), status=500)


@web.middleware
async def handle_auth(request, handler):
    if request.match_info.http_exception is not None:
        raise request.match_info.http_exception

    authorization_info = app.route_auth_mapping.get(request.match_info.route)
    method = request.method
    path = request.match_info.route.resource.canonical

    request = authentication.process_request(request)

    # Only go through authentication if route specifies permissions;
    # otherwise route is considered public.
    if authorization_info is not None:
        try:
            account = authentication.authenticate(request)

        except authentication.AuthenticationException:
            raise api.UnauthenticatedException()

        if not authorization.authorize(account, method, path, authorization_info):
            raise api.ForbiddenException()

    response = await handler(request)
    response = authentication.process_response(request, response)

    return response


@web.middleware
async def handle_schema_validation(request, handler):
    return await validation_middleware(request, handler)


@aiohttpparser.parser.error_handler
def _handle_schema_validation_error(error, req, schema, status_code, headers):
    raise HTTPSchemaValidationError(error)





class AuthenticationException(Exception):
    pass


class NoSuchAccount(AuthenticationException):
    def __init__(self):
        super().__init__('no such account')


class IdentityVerificationFailed(AuthenticationException):
    def __init__(self):
        super().__init__('identity verification failed')




from colibris.authorization import ANY_PERMISSION


class AuthorizationBackend:
    def authorize(self, account, method, path, required_permissions):
        # by default, when "any" permission is fine,
        # consider any authenticated request authorized

        return required_permissions == ANY_PERMISSION


class NullBackend(AuthorizationBackend):
    def __init__(self, **kwargs):
        pass

    def authorize(self, account, method, path, required_permissions):
        # consider any authenticated request authorized

        return True




import logging

from colibris import settings
from colibris import utils


ANY_PERMISSION = '*'

logger = logging.getLogger(__name__)

_backend = None


def get_backend():
    global _backend

    if _backend is None:
        backend_settings = dict(settings.AUTHORIZATION)
        backend_path = backend_settings.pop('backend', 'colibris.authorization.base.NullBackend')
        backend_class = utils.import_member(backend_path)

        logger.debug('creating backend of class %s', backend_path)

        _backend = backend_class(**backend_settings)

    return _backend


def authorize(account, method, path, authorization_info):
    return get_backend().authorize(account, method, path, authorization_info)
